"""Tests with the whole logger disabled"""

from logbook import Logger

log = Logger("Test logger")
log.disabled = True


def run():
    for x in range(500):
        log.warning("this is not handled")



import importlib

import pytest


@pytest.fixture(params=["speedups", "fallback"])
def speedups_module(request):
    mod_name = f"logbook._{request.param}"
    try:
        return importlib.import_module(mod_name)
    except ImportError:
        pytest.skip(f"{mod_name} is not available")


def test_group_reflected_property(speedups_module):
    class Group:
        foo = "group"

    class A:
        foo = speedups_module.group_reflected_property("foo", "default")

        def __init__(self, group=None):
            self.group = group

    a = A()
    assert a.foo == "default"
    a.group = Group()
    assert a.foo == "group"
    a.foo = "set"
    assert a.foo == "set"
    del a.foo
    assert a.foo == "group"



import functools
import re
import warnings
from io import StringIO
from random import randrange

import pytest

import logbook
import logbook.compat

from .utils import capturing_stderr_context

__file_without_pyc__ = __file__
if __file_without_pyc__.endswith(".pyc"):
    __file_without_pyc__ = __file_without_pyc__[:-1]


@pytest.mark.parametrize("set_root_logger_level", [True, False])
def test_basic_compat(request, set_root_logger_level):
    import logging

    from logbook.compat import redirected_logging

    # mimic the default logging setting
    request.addfinalizer(functools.partial(logging.root.setLevel, logging.root.level))
    logging.root.setLevel(logging.WARNING)

    name = "test_logbook-%d" % randrange(1 << 32)
    logger = logging.getLogger(name)

    with logbook.TestHandler(bubble=True) as handler:
        with capturing_stderr_context() as captured:
            with redirected_logging(set_root_logger_level):
                logger.debug("This is from the old system")
                logger.info("This is from the old system")
                logger.warning("This is from the old %s", "system")
                logger.error("This is from the old system")
                logger.critical("This is from the old system")
                logger.error(
                    "This is a %(what)s %(where)s", {"what": "mapping", "where": "test"}
                )
        assert (
            "WARNING: %s: This is from the old system" % name
        ) in captured.getvalue()
        assert ("ERROR: %s: This is a mapping test" % name) in captured.getvalue()
    if set_root_logger_level:
        assert handler.records[0].level == logbook.DEBUG
    else:
        assert handler.records[0].level == logbook.WARNING
        assert handler.records[0].msg == "This is from the old %s"


def test_redirect_logbook():
    import logging

    out = StringIO()
    logger = logging.getLogger()
    logbook_logger = logbook.Logger("testlogger")
    old_handlers = logger.handlers[:]
    handler = logging.StreamHandler(out)
    handler.setFormatter(logging.Formatter("%(name)s:%(levelname)s:%(message)s"))
    logger.handlers[:] = [handler]
    try:
        with logbook.compat.LoggingHandler():
            logbook_logger.warn("This goes to logging")
            pieces = out.getvalue().strip().split(":")
            assert pieces == ["testlogger", "WARNING", "This goes to logging"]
    finally:
        logger.handlers[:] = old_handlers


from itertools import count

test_warning_redirections_i = count()


def test_warning_redirections():
    from logbook.compat import redirected_warnings

    with logbook.TestHandler() as handler:
        with redirected_warnings():
            warnings.warn(
                RuntimeWarning(f"Testing {next(test_warning_redirections_i)}")
            )

    assert len(handler.formatted_records) == 1
    assert handler.formatted_records[0].startswith("[WARNING] RuntimeWarning: Testing")
    assert __file_without_pyc__ in handler.records[0].filename



from datetime import datetime

import pytest


def test_jsonhelper():
    from logbook.helpers import to_safe_json

    class Bogus:
        def __str__(self):
            return "bogus"

    rv = to_safe_json(
        [
            None,
            "foo",
            "jäger",
            1,
            datetime(2000, 1, 1),
            {"jäger1": 1, "jäger2": 2, Bogus(): 3, "invalid": object()},
            object(),  # invalid
        ]
    )

    assert rv == [
        None,
        "foo",
        "jäger",
        1,
        "2000-01-01T00:00:00Z",
        {"jäger1": 1, "jäger2": 2, "bogus": 3, "invalid": None},
        None,
    ]


def test_datehelpers():
    from logbook.helpers import format_iso8601, parse_iso8601

    now = datetime.now()
    rv = format_iso8601()
    assert rv[:4] == str(now.year)

    with pytest.raises(ValueError):
        parse_iso8601("foo")
    v = parse_iso8601("2000-01-01T00:00:00.12Z")
    assert v.microsecond == 120000
    v = parse_iso8601("2000-01-01T12:00:00+01:00")
    assert v.hour == 11
    v = parse_iso8601("2000-01-01T12:00:00-01:00")
    assert v.hour == 13



"""
    test utils for logbook
    ~~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
    :license: BSD, see LICENSE for more details.
"""

import functools
import importlib
import sys
from contextlib import contextmanager
from io import StringIO

import pytest

import logbook

_missing = object()

LETTERS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"


def require_module(module_name):
    found = True
    try:
        importlib.import_module(module_name)
    except ImportError:
        found = False

    return pytest.mark.skipif(not found, reason=f"Module {module_name} is required")


def make_fake_mail_handler(**kwargs):
    class FakeMailHandler(logbook.MailHandler):
        mails = []

        def get_connection(self):
            return self

        def close_connection(self, con):
            pass

        def sendmail(self, fromaddr, recipients, mail):
            self.mails.append((fromaddr, recipients, mail))

    kwargs.setdefault("level", logbook.ERROR)
    return FakeMailHandler("foo@example.com", ["bar@example.com"], **kwargs)


def missing(name):
    def decorate(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            old = sys.modules.get(name, _missing)
            sys.modules[name] = None
            try:
                f(*args, **kwargs)
            finally:
                if old is _missing:
                    del sys.modules[name]
                else:
                    sys.modules[name] = old

        return wrapper

    return decorate


def activate_via_with_statement(handler):
    return handler


@contextmanager
def activate_via_push_pop(handler):
    handler.push_thread()
    try:
        yield handler
    finally:
        handler.pop_thread()


@contextmanager
def capturing_stderr_context():
    original = sys.stderr
    sys.stderr = StringIO()
    try:
        yield sys.stderr
    finally:
        sys.stderr = original



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from importlib.metadata import distribution

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Logbook"
project_copyright = "2010, Armin Ronacher, Georg Brandl"
version = release = distribution("Logbook").version


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sheet"
html_theme_options = {
    "nosidebar": True,
}
html_theme_path = ["."]
html_title = "Logbook"
html_short_title = "Logbook " + release
html_static_path = ["_static"]


# -- Extension configuration -------------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}



import sys
from io import StringIO

import pytest

import logbook

from .utils import capturing_stderr_context, missing, require_module


@require_module("jinja2")
def test_jinja_formatter(logger):
    from logbook.more import JinjaFormatter

    fmter = JinjaFormatter("{{ record.channel }}/{{ record.level_name }}")
    handler = logbook.TestHandler()
    handler.formatter = fmter
    with handler:
        logger.info("info")
    assert "testlogger/INFO" in handler.formatted_records


@missing("jinja2")
def test_missing_jinja2():
    from logbook.more import JinjaFormatter

    # check the RuntimeError is raised
    with pytest.raises(RuntimeError):
        JinjaFormatter("dummy")


def test_colorizing_support(logger):
    from logbook.more import ColorizedStderrHandler

    class TestColorizingHandler(ColorizedStderrHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._obj_stream = StringIO()

        @property
        def stream(self):
            return self._obj_stream

    with TestColorizingHandler(format_string="{record.message}") as handler:
        handler.force_color()
        logger.error("An error")
        logger.warn("A warning")
        logger.debug("A debug message")
        lines = handler.stream.getvalue().rstrip("\n").splitlines()
        assert lines == [
            "\x1b[31;01mAn error\x1b[39;49;00m",
            "\x1b[33;01mA warning\x1b[39;49;00m",
            "\x1b[37mA debug message\x1b[39;49;00m",
        ]

    with TestColorizingHandler(format_string="{record.message}") as handler:
        handler.forbid_color()
        logger.error("An error")
        logger.warn("A warning")
        logger.debug("A debug message")
        lines = handler.stream.getvalue().rstrip("\n").splitlines()
        assert lines == ["An error", "A warning", "A debug message"]


def test_tagged(default_handler):
    from logbook.more import TaggingHandler, TaggingLogger

    stream = StringIO()
    second_handler = logbook.StreamHandler(stream)

    logger = TaggingLogger("name", ["cmd"])
    handler = TaggingHandler(
        dict(
            info=default_handler,
            cmd=second_handler,
            both=[default_handler, second_handler],
        )
    )
    handler.bubble = False

    with handler:
        with capturing_stderr_context() as captured:
            logger.log("info", "info message")
            logger.log("both", "all message")
            logger.cmd("cmd message")

    stderr = captured.getvalue()

    assert "info message" in stderr
    assert "all message" in stderr
    assert "cmd message" not in stderr

    stringio = stream.getvalue()

    assert "info message" not in stringio
    assert "all message" in stringio
    assert "cmd message" in stringio


def test_tagging_logger(default_handler):
    from logbook import StderrHandler
    from logbook.more import TaggingLogger

    logger = TaggingLogger("tagged", ["a", "b"])
    handler = StderrHandler(format_string="{record.msg}|{record.extra[tags]}")

    with handler:
        with capturing_stderr_context() as captured:
            logger.a("a")
            logger.b("b")

    stderr = captured.getvalue()

    assert "a|['a']" in stderr
    assert "a|['b']" not in stderr
    assert "b|['b']" in stderr
    assert "b|['a']" not in stderr


def test_external_application_handler(tmpdir, logger):
    from logbook.more import ExternalApplicationHandler as Handler

    fn = tmpdir.join("tempfile")
    handler = Handler(
        [
            sys.executable,
            "-c",
            r"""if 1:
    f = open({tempfile}, 'w')
    try:
        f.write('{{record.message}}\n')
    finally:
        f.close()
    """.format(
                tempfile=repr(str(fn))
            ),
        ]
    )
    with handler:
        logger.error("this is a really bad idea")
    with fn.open() as rf:
        contents = rf.read().strip()
    assert contents == "this is a really bad idea"


def test_exception_handler(logger):
    from logbook.more import ExceptionHandler

    with ExceptionHandler(ValueError):
        with pytest.raises(ValueError) as caught:
            logger.info("here i am")
    assert "INFO: testlogger: here i am" in caught.value.args[0]


def test_exception_handler_specific_level(logger):
    from logbook.more import ExceptionHandler

    with logbook.TestHandler() as test_handler:
        with pytest.raises(ValueError) as caught:
            with ExceptionHandler(ValueError, level="WARNING"):
                logger.info("this is irrelevant")
                logger.warn("here i am")
        assert "WARNING: testlogger: here i am" in caught.value.args[0]
    assert "this is irrelevant" in test_handler.records[0].message


def test_dedup_handler(logger):
    from logbook.more import DedupHandler

    with logbook.TestHandler() as test_handler:
        with DedupHandler():
            logger.info("foo")
            logger.info("bar")
            logger.info("foo")
    assert 2 == len(test_handler.records)
    assert "message repeated 2 times: foo" in test_handler.records[0].message
    assert "message repeated 1 times: bar" in test_handler.records[1].message


class TestRiemannHandler:
    @require_module("riemann_client")
    def test_happy_path(self, logger):
        from logbook.more import RiemannHandler

        riemann_handler = RiemannHandler(
            "127.0.0.1", 5555, message_type="test", level=logbook.INFO
        )
        null_handler = logbook.NullHandler()
        with null_handler.applicationbound():
            with riemann_handler:
                logger.error("Something bad has happened")
                try:
                    raise RuntimeError("For example, a RuntimeError")
                except Exception as ex:
                    logger.exception(ex)
                logger.info("But now it is ok")

        q = riemann_handler.queue
        assert len(q) == 3
        error_event = q[0]
        assert error_event["state"] == "error"
        exc_event = q[1]
        assert exc_event["description"] == "For example, a RuntimeError"
        info_event = q[2]
        assert info_event["state"] == "ok"

    @require_module("riemann_client")
    def test_incorrect_type(self):
        from logbook.more import RiemannHandler

        with pytest.raises(RuntimeError):
            RiemannHandler("127.0.0.1", 5555, message_type="fancy_type")

    @require_module("riemann_client")
    def test_flush(self, logger):
        from logbook.more import RiemannHandler

        riemann_handler = RiemannHandler(
            "127.0.0.1",
            5555,
            message_type="test",
            flush_threshold=2,
            level=logbook.INFO,
        )
        null_handler = logbook.NullHandler()
        with null_handler.applicationbound():
            with riemann_handler:
                logger.info("Msg #1")
                logger.info("Msg #2")
                logger.info("Msg #3")

        q = riemann_handler.queue
        assert len(q) == 1
        assert q[0]["description"] == "Msg #3"



import re
import sys

import pytest

import logbook

from .utils import capturing_stderr_context

__file_without_pyc__ = __file__
if __file_without_pyc__.endswith(".pyc"):
    __file_without_pyc__ = __file_without_pyc__[:-1]


def test_handler_exception(activation_strategy, logger):
    class ErroringHandler(logbook.TestHandler):
        def emit(self, record):
            raise RuntimeError("something bad happened")

    with capturing_stderr_context() as stderr:
        with activation_strategy(ErroringHandler()):
            logger.warn("I warn you.")
    assert "something bad happened" in stderr.getvalue()
    assert "I warn you" not in stderr.getvalue()


def test_formatting_exception():
    def make_record():
        return logbook.LogRecord(
            "Test Logger",
            logbook.WARNING,
            "Hello {foo:invalid}",
            kwargs={"foo": 42},
            frame=sys._getframe(),
        )

    record = make_record()
    with pytest.raises(TypeError) as caught:
        record.message

    errormsg = str(caught.value)
    assert re.search(
        "Could not format message with provided arguments: Invalid "
        "(?:format specifier)|(?:conversion specification)|(?:format spec)",
        errormsg,
        re.M | re.S,
    )
    assert "msg='Hello {foo:invalid}'" in errormsg
    assert "args=()" in errormsg
    assert "kwargs={'foo': 42}" in errormsg
    assert re.search(
        r"Happened in file .*%s, line \d+" % re.escape(__file_without_pyc__),
        errormsg,
        re.M | re.S,
    )



"""Tests with a logging handler becoming a noop for comparison"""

from io import StringIO
from logging import ERROR, StreamHandler, getLogger

log = getLogger("Testlogger")
log.setLevel(ERROR)


def run():
    out = StringIO()
    handler = StreamHandler(out)
    log.addHandler(handler)
    for x in range(500):
        log.warning("this is not handled")



from pathlib import Path

import pytest

import logbook

logbook.StderrHandler().push_application()


@pytest.fixture
def logger():
    return logbook.Logger("testlogger")


@pytest.fixture
def active_handler(request, test_handler, activation_strategy):
    s = activation_strategy(test_handler)
    s.activate()

    @request.addfinalizer
    def deactivate():
        s.deactivate()

    return test_handler


@pytest.fixture
def test_handler():
    return logbook.TestHandler()


class ActivationStrategy:
    def __init__(self, handler):
        super().__init__()
        self.handler = handler

    def activate(self):
        raise NotImplementedError()  # pragma: no cover

    def deactivate(self):
        raise NotImplementedError()  # pragma: no cover

    def __enter__(self):
        self.activate()
        return self.handler

    def __exit__(self, *_):
        self.deactivate()


class ContextEnteringStrategy(ActivationStrategy):
    def activate(self):
        self.handler.__enter__()

    def deactivate(self):
        self.handler.__exit__(None, None, None)


class PushingStrategy(ActivationStrategy):
    def activate(self):
        from logbook.concurrency import is_gevent_enabled

        if is_gevent_enabled():
            self.handler.push_greenlet()
        else:
            self.handler.push_thread()

    def deactivate(self):
        from logbook.concurrency import is_gevent_enabled

        if is_gevent_enabled():
            self.handler.pop_greenlet()
        else:
            self.handler.pop_thread()


@pytest.fixture(params=[ContextEnteringStrategy, PushingStrategy])
def activation_strategy(request):
    return request.param


class CustomPathLike:
    def __init__(self, path):
        self.path = path

    def __fspath__(self):
        return self.path


@pytest.fixture(params=[Path, str, CustomPathLike])
def logfile(tmp_path, request):
    path = str(tmp_path / "logfile.log")
    return request.param(path)


@pytest.fixture
def default_handler(request):
    returned = logbook.StderrHandler()
    returned.push_application()
    request.addfinalizer(returned.pop_application)
    return returned


try:
    import gevent
except ImportError:
    pass
else:

    @pytest.fixture(
        scope="module", autouse=True, params=[False, True], ids=["nogevent", "gevent"]
    )
    def gevent(request):
        module_name = getattr(request.module, "__name__", "")
        if (
            not any(s in module_name for s in ("queues", "processors"))
            and request.param
        ):
            from logbook.concurrency import _disable_gevent, enable_gevent

            enable_gevent()

            @request.addfinalizer
            def fin():
                _disable_gevent()



"""Tests with a filter disabling a handler for comparsion in logging"""

from io import StringIO
from logging import Filter, StreamHandler, getLogger

log = getLogger("Testlogger")


class DisableFilter(Filter):
    def filter(self, record):
        return False


def run():
    out = StringIO()
    handler = StreamHandler(out)
    handler.addFilter(DisableFilter())
    log.addHandler(handler)
    for x in range(500):
        log.warning("this is not handled")



"""Tests the stream handler in logging"""

from io import StringIO
from logging import Logger, StreamHandler

log = Logger("Test logger")


def run():
    out = StringIO()
    log.addHandler(StreamHandler(out))
    for x in range(500):
        log.warning("this is not handled")
    assert out.getvalue().count("\n") == 500



"""
    logbook.helpers
    ~~~~~~~~~~~~~~~

    Various helper functions

    :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
    :license: BSD, see LICENSE for more details.
"""

import errno
import os
import random
import re
import sys
import time
from datetime import datetime, timedelta, timezone

# this regexp also matches incompatible dates like 20070101 because
# some libraries (like the python xmlrpclib modules) use this
_iso8601_re = re.compile(
    # date
    r"(\d{4})(?:-?(\d{2})(?:-?(\d{2}))?)?"
    # time
    r"(?:T(\d{2}):(\d{2})(?::(\d{2}(?:\.\d+)?))?(Z|[+-]\d{2}:\d{2})?)?$"
)
_missing = object()


can_rename_open_file = False
if os.name == "nt":
    try:
        import ctypes

        _MOVEFILE_REPLACE_EXISTING = 0x1
        _MOVEFILE_WRITE_THROUGH = 0x8
        _MoveFileEx = ctypes.windll.kernel32.MoveFileExW

        def _rename(src, dst):
            if _rename_atomic(src, dst):
                return True
            retry = 0
            rv = False
            while not rv and retry < 100:
                rv = _MoveFileEx(
                    src, dst, _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH
                )
                if not rv:
                    time.sleep(0.001)
                    retry += 1
            return rv

        # new in Vista and Windows Server 2008
        _CreateTransaction = ctypes.windll.ktmw32.CreateTransaction
        _CommitTransaction = ctypes.windll.ktmw32.CommitTransaction
        _MoveFileTransacted = ctypes.windll.kernel32.MoveFileTransactedW
        _CloseHandle = ctypes.windll.kernel32.CloseHandle
        can_rename_open_file = True

        def _rename_atomic(src, dst):
            ta = _CreateTransaction(None, 0, 0, 0, 0, 1000, "Logbook rename")
            if ta == -1:
                return False
            try:
                retry = 0
                rv = False
                while not rv and retry < 100:
                    if rv := _MoveFileTransacted(
                        src,
                        dst,
                        None,
                        None,
                        _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH,
                        ta,
                    ):
                        rv = _CommitTransaction(ta)
                        break
                    else:
                        time.sleep(0.001)
                        retry += 1
                return rv
            finally:
                _CloseHandle(ta)

    except Exception:

        def _rename(src, dst):
            return False

        def _rename_atomic(src, dst):
            return False

    def rename(src, dst):
        # Try atomic or pseudo-atomic rename
        if _rename(src, dst):
            return
        # Fall back to "move away and replace"
        try:
            os.rename(src, dst)
        except OSError:
            e = sys.exc_info()[1]
            if e.errno not in (errno.EEXIST, errno.EACCES):
                raise
            old = f"{dst}-{random.randint(0, 2**31 - 1):08x}"
            os.rename(dst, old)
            os.rename(src, dst)
            try:
                os.unlink(old)
            except Exception:
                pass

else:
    rename = os.rename
    can_rename_open_file = True

_JSON_SIMPLE_TYPES = (bool, float, int, str)


def to_safe_json(data):
    """Makes a data structure safe for JSON silently discarding invalid
    objects from nested structures.  This also converts dates.
    """

    def _convert(obj):
        if obj is None:
            return None
        elif isinstance(obj, _JSON_SIMPLE_TYPES):
            return obj
        elif isinstance(obj, datetime):
            return format_iso8601(obj)
        elif isinstance(obj, list):
            return [_convert(x) for x in obj]
        elif isinstance(obj, tuple):
            return tuple(_convert(x) for x in obj)
        elif isinstance(obj, dict):
            rv = {}
            for key, value in obj.items():
                if not isinstance(key, str):
                    key = str(key)
                rv[key] = _convert(value)
            return rv

    return _convert(data)


if sys.version_info >= (3, 12):

    def datetime_utcnow():
        """datetime.utcnow() but doesn't emit a deprecation warning.

        Will be fixed by https://github.com/getlogbook/logbook/issues/353
        """
        return datetime.now(timezone.utc).replace(tzinfo=None)

    def datetime_utcfromtimestamp(timestamp):
        """datetime.utcfromtimesetamp() but doesn't emit a deprecation warning.

        Will be fixed by https://github.com/getlogbook/logbook/issues/353
        """
        return datetime.fromtimestamp(timestamp, timezone.utc).replace(tzinfo=None)

else:
    datetime_utcnow = datetime.utcnow
    datetime_utcfromtimestamp = datetime.utcfromtimestamp


def format_iso8601(d=None):
    """Returns a date in iso8601 format."""
    if d is None:
        d = datetime_utcnow()
    rv = d.strftime("%Y-%m-%dT%H:%M:%S")
    if d.microsecond:
        rv += "." + str(d.microsecond)
    return rv + "Z"


def parse_iso8601(value):
    """Parse an iso8601 date into a datetime object.  The timezone is
    normalized to UTC.
    """
    m = _iso8601_re.match(value)
    if m is None:
        raise ValueError("not a valid iso8601 date value")

    groups = m.groups()
    args = []
    for group in groups[:-2]:
        if group is not None:
            group = int(group)
        args.append(group)
    seconds = groups[-2]
    if seconds is not None:
        if "." in seconds:
            sec, usec = seconds.split(".")
            args.append(int(sec))
            args.append(int(usec.ljust(6, "0")))
        else:
            args.append(int(seconds))

    rv = datetime(*args)
    tz = groups[-1]
    if tz and tz != "Z":
        args = [int(x) for x in tz[1:].split(":")]
        delta = timedelta(hours=args[0], minutes=args[1])
        if tz[0] == "+":
            rv -= delta
        else:
            rv += delta

    return rv


def get_application_name():
    if not sys.argv or not sys.argv[0]:
        return "Python"
    return os.path.basename(sys.argv[0]).title()


class cached_property:
    """A property that is lazily calculated and then cached."""

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def get_iterator_next_method(it):
    return lambda: next(it)



import pytest

import logbook

from .utils import capturing_stderr_context


def test_error_flag(logger):
    with capturing_stderr_context() as captured:
        with logbook.Flags(errors="print"):
            with logbook.Flags(errors="silent"):
                logger.warn("Foo {42}", "aha")
        assert captured.getvalue() == ""

        with logbook.Flags(errors="silent"):
            with logbook.Flags(errors="print"):
                logger.warn("Foo {42}", "aha")
        assert captured.getvalue() != ""

        with pytest.raises(Exception) as caught:
            with logbook.Flags(errors="raise"):
                logger.warn("Foo {42}", "aha")
        assert "Could not format message with provided arguments" in str(caught.value)


def test_disable_introspection(logger):
    with logbook.Flags(introspection=False):
        with logbook.TestHandler() as h:
            logger.warn("Testing")
            assert h.records[0].frame is None
            assert h.records[0].calling_frame is None
            assert h.records[0].module is None



import sys

import logbook

from .utils import capturing_stderr_context


def test_exc_info_when_no_exceptions_exist(logger):
    with capturing_stderr_context() as captured:
        with logbook.StreamHandler(sys.stderr):
            logger.debug("message", exc_info=True)
    assert "Traceback" not in captured.getvalue()


def test_exc_info_false():
    with logbook.handlers.TestHandler() as handler:
        logbook.debug("message here", exc_info=False)
    [record] = handler.records
    assert not record.formatted_exception


def test_exc_info_exception_instance(logger):
    with logbook.handlers.TestHandler() as handler:
        try:
            raise ValueError("error here")
        except Exception as e:
            error = e
        logger.exception(exc_info=error)
    [record] = handler.records
    assert isinstance(record.exc_info, tuple)
    assert len(record.exc_info) == 3
    assert "Traceback" in record.formatted_exception


def test_extradict(active_handler, logger):
    logger.warn("Test warning")
    record = active_handler.records[0]
    record.extra["existing"] = "foo"
    assert record.extra["nonexisting"] == ""
    assert record.extra["existing"] == "foo"


def test_calling_frame(active_handler, logger):
    logger.warn("test")
    assert active_handler.records[0].calling_frame == sys._getframe()


def test_frame_correction(active_handler, logger):
    def inner():
        logger.warn("test", frame_correction=+1)

    inner()
    assert active_handler.records[0].calling_frame == sys._getframe()


def test_dispatcher(active_handler, logger):
    logger.warn("Logbook is too awesome for stdlib")
    assert active_handler.records[0].dispatcher == logger



import pickle
import sys

import pytest

import logbook


def test_basic_logging(active_handler, logger):
    logger.warn("This is a warning.  Nice hah?")

    assert active_handler.has_warning("This is a warning.  Nice hah?")
    assert active_handler.formatted_records == [
        "[WARNING] testlogger: This is a warning.  Nice hah?"
    ]


def test_exception_catching(active_handler, logger):
    assert not active_handler.has_error()
    try:
        1 / 0
    except Exception:
        logger.exception()
    try:
        1 / 0
    except Exception:
        logger.exception("Awesome")
    assert active_handler.has_error("Uncaught exception occurred")
    assert active_handler.has_error("Awesome")
    assert active_handler.records[0].exc_info is not None
    assert "1 / 0" in active_handler.records[0].formatted_exception


def test_exception_catching_with_unicode():
    """See https://github.com/getlogbook/logbook/issues/104"""
    try:
        raise Exception("\u202a test \u202c")
    except:
        r = logbook.LogRecord("channel", "DEBUG", "test", exc_info=sys.exc_info())
    r.exception_message


@pytest.mark.parametrize("as_tuple", [True, False])
def test_exc_info(as_tuple, logger, active_handler):
    try:
        1 / 0
    except Exception:
        exc_info = sys.exc_info()
        logger.info("Exception caught", exc_info=exc_info if as_tuple else True)
    assert active_handler.records[0].exc_info is not None
    assert active_handler.records[0].exc_info == exc_info


def test_to_dict(logger, active_handler):
    try:
        1 / 0
    except Exception:
        logger.exception()
        record = active_handler.records[0]

    exported = record.to_dict()
    record.close()
    imported = logbook.LogRecord.from_dict(exported)
    for key, value in record.__dict__.items():
        if key[0] == "_":
            continue
        assert value == getattr(imported, key)


def test_pickle(active_handler, logger):
    try:
        1 / 0
    except Exception:
        logger.exception()
        record = active_handler.records[0]
    record.pull_information()
    record.close()

    for p in range(pickle.HIGHEST_PROTOCOL):
        exported = pickle.dumps(record, p)
        imported = pickle.loads(exported)
        for key, value in record.__dict__.items():
            if key[0] == "_":
                continue
            imported_value = getattr(imported, key)
            if isinstance(value, ZeroDivisionError):
                # in Python 3.2, ZeroDivisionError(x) != ZeroDivisionError(x)
                assert type(value) is type(imported_value)
                assert value.args == imported_value.args
            else:
                assert value == imported_value



import logbook

from .utils import capturing_stderr_context, make_fake_mail_handler


def test_custom_logger(activation_strategy, logger):
    client_ip = "127.0.0.1"

    class CustomLogger(logbook.Logger):
        def process_record(self, record):
            record.extra["ip"] = client_ip

    custom_log = CustomLogger("awesome logger")
    fmt = (
        "[{record.level_name}] {record.channel}: "
        "{record.message} [{record.extra[ip]}]"
    )
    handler = logbook.TestHandler(format_string=fmt)
    assert handler.format_string == fmt

    with activation_strategy(handler):
        custom_log.warn("Too many sounds")
        logger.warn('"Music" playing')

    assert handler.formatted_records == [
        "[WARNING] awesome logger: Too many sounds [127.0.0.1]",
        '[WARNING] testlogger: "Music" playing []',
    ]


def test_custom_handling(activation_strategy, logger):
    class MyTestHandler(logbook.TestHandler):
        def handle(self, record):
            if record.extra.get("flag") != "testing":
                return False
            return logbook.TestHandler.handle(self, record)

    # Check metaclass (== cls.__class__)
    assert logbook.Handler.__class__ == logbook.handlers._HandlerType

    class MyLogger(logbook.Logger):
        def process_record(self, record):
            logbook.Logger.process_record(self, record)
            record.extra["flag"] = "testing"

    log = MyLogger()
    handler = MyTestHandler()
    with capturing_stderr_context() as captured:
        with activation_strategy(handler):
            log.warn("From my logger")
            logger.warn("From another logger")
        assert handler.has_warning("From my logger")
        assert "From another logger" in captured.getvalue()


def test_nested_setups(activation_strategy):
    with capturing_stderr_context() as captured:
        logger = logbook.Logger("App")
        test_handler = logbook.TestHandler(level="WARNING")
        mail_handler = make_fake_mail_handler(bubble=True)

        handlers = logbook.NestedSetup(
            [logbook.NullHandler(), test_handler, mail_handler]
        )

        with activation_strategy(handlers):
            logger.warn("This is a warning")
            logger.error("This is also a mail")
            try:
                1 / 0
            except Exception:
                logger.exception()
        logger.warn("And here we go straight back to stderr")

        assert test_handler.has_warning("This is a warning")
        assert test_handler.has_error("This is also a mail")
        assert len(mail_handler.mails) == 2
        assert "This is also a mail" in mail_handler.mails[0][2]
        assert "1 / 0" in mail_handler.mails[1][2]
        assert "And here we go straight back to stderr" in captured.getvalue()

        with activation_strategy(handlers):
            logger.warn("threadbound warning")

        handlers.push_application()
        try:
            logger.warn("applicationbound warning")
        finally:
            handlers.pop_application()


def test_filtering(activation_strategy):
    logger1 = logbook.Logger("Logger1")
    logger2 = logbook.Logger("Logger2")
    handler = logbook.TestHandler()
    outer_handler = logbook.TestHandler()

    def only_1(record, handler):
        return record.dispatcher is logger1

    handler.filter = only_1

    with activation_strategy(outer_handler):
        with activation_strategy(handler):
            logger1.warn("foo")
            logger2.warn("bar")

    assert handler.has_warning("foo", channel="Logger1")
    assert not handler.has_warning("bar", channel="Logger2")
    assert not outer_handler.has_warning("foo", channel="Logger1")
    assert outer_handler.has_warning("bar", channel="Logger2")


def test_different_context_pushing(activation_strategy):
    h1 = logbook.TestHandler(level=logbook.DEBUG)
    h2 = logbook.TestHandler(level=logbook.INFO)
    h3 = logbook.TestHandler(level=logbook.WARNING)
    logger = logbook.Logger("Testing")

    with activation_strategy(h1):
        with activation_strategy(h2):
            with activation_strategy(h3):
                logger.warn("Wuuu")
                logger.info("still awesome")
                logger.debug("puzzled")

    assert h1.has_debug("puzzled")
    assert h2.has_info("still awesome")
    assert h3.has_warning("Wuuu")
    for handler in h1, h2, h3:
        assert len(handler.records) == 1


def test_default_handlers(logger):
    with capturing_stderr_context() as stream:
        logger.warn("Aha!")
        captured = stream.getvalue()
    assert "WARNING: testlogger: Aha!" in captured



import asyncio

import logbook

ITERATIONS = 100


def test_asyncio_context_management(logger):
    h1 = logbook.TestHandler()
    h2 = logbook.TestHandler()

    async def task(handler, msg):
        for _ in range(ITERATIONS):
            with handler.contextbound():
                logger.info(msg)

            await asyncio.sleep(0)  # allow for context switch

    async def main():
        await asyncio.gather(task(h1, "task1"), task(h2, "task2"))

    asyncio.run(main())

    assert len(h1.records) == ITERATIONS
    assert all(["task1" == r.msg for r in h1.records])

    assert len(h2.records) == ITERATIONS
    assert all(["task2" == r.msg for r in h2.records])



from time import sleep
from unittest.mock import Mock, call

import pytest

import logbook
from logbook.utils import (
    deprecated,
    forget_deprecation_locations,
    log_deprecation_message,
    logged_if_slow,
    suppressed_deprecations,
)

_THRESHOLD = 0.1


@pytest.mark.flaky(reruns=5)
def test_logged_if_slow_reached(test_handler):
    with test_handler.applicationbound():
        with logged_if_slow("checking...", threshold=_THRESHOLD):
            sleep(2 * _THRESHOLD)
        assert len(test_handler.records) == 1
        [record] = test_handler.records
        assert record.message == "checking..."


@pytest.mark.flaky(reruns=5)
def test_logged_if_slow_did_not_reached(test_handler):
    with test_handler.applicationbound():
        with logged_if_slow("checking...", threshold=_THRESHOLD):
            sleep(_THRESHOLD / 2)
        assert len(test_handler.records) == 0


@pytest.mark.flaky(reruns=5)
def test_logged_if_slow_logger():
    logger = Mock()

    with logged_if_slow("checking...", threshold=_THRESHOLD, logger=logger):
        sleep(2 * _THRESHOLD)

    assert logger.log.call_args == call(logbook.DEBUG, "checking...")


@pytest.mark.flaky(reruns=5)
def test_logged_if_slow_level(test_handler):
    with test_handler.applicationbound():
        with logged_if_slow("checking...", threshold=_THRESHOLD, level=logbook.WARNING):
            sleep(2 * _THRESHOLD)

    assert test_handler.records[0].level == logbook.WARNING


@pytest.mark.flaky(reruns=5)
def test_logged_if_slow_deprecated(logger, test_handler):
    with test_handler.applicationbound():
        with logged_if_slow("checking...", threshold=_THRESHOLD, func=logbook.error):
            sleep(2 * _THRESHOLD)

    assert test_handler.records[0].level == logbook.ERROR
    assert test_handler.records[0].message == "checking..."

    with pytest.raises(TypeError):
        logged_if_slow("checking...", logger=logger, func=logger.error)


def test_deprecated_func_called(capture):
    assert deprecated_func(1, 2) == 3


def test_deprecation_message(capture):
    deprecated_func(1, 2)

    [record] = capture.records
    assert "deprecated" in record.message
    assert "deprecated_func" in record.message


def test_deprecation_with_message(capture):
    @deprecated("use something else instead")
    def func(a, b):
        return a + b

    func(1, 2)

    [record] = capture.records
    assert "use something else instead" in record.message
    assert "func is deprecated" in record.message


def test_no_deprecations(capture):
    @deprecated("msg")
    def func(a, b):
        return a + b

    with suppressed_deprecations():
        assert func(1, 2) == 3
    assert not capture.records


def _no_decorator(func):
    return func


@pytest.mark.parametrize("decorator", [_no_decorator, classmethod])
def test_class_deprecation(capture, decorator):
    class Bla:
        @deprecated("reason")
        @classmethod
        def func(self, a, b):
            assert isinstance(self, Bla)
            return a + b

    assert Bla().func(2, 4) == 6

    [record] = capture.records
    assert "Bla.func is deprecated" in record.message


def test_deprecations_different_sources(capture):
    def f():
        deprecated_func(1, 2)

    def g():
        deprecated_func(1, 2)

    f()
    g()
    assert len(capture.records) == 2


def test_deprecations_same_sources(capture):
    def f():
        deprecated_func(1, 2)

    f()
    f()
    assert len(capture.records) == 1


def test_deprecation_message_different_sources(capture):
    def f(flag):
        if flag:
            log_deprecation_message("first message type")
        else:
            log_deprecation_message("second message type")

    f(True)
    f(False)
    assert len(capture.records) == 2


def test_deprecation_message_same_sources(capture):
    def f(flag):
        if flag:
            log_deprecation_message("first message type")
        else:
            log_deprecation_message("second message type")

    f(True)
    f(True)
    assert len(capture.records) == 1


def test_deprecation_message_full_warning(capture):
    def f():
        log_deprecation_message("some_message")

    f()

    [record] = capture.records
    assert record.message == "Deprecation message: some_message"


def test_name_doc():
    @deprecated
    def some_func():
        """docstring here"""
        pass

    assert some_func.__name__ == "some_func"
    assert "docstring here" in some_func.__doc__


def test_doc_update():
    @deprecated("some_message")
    def some_func():
        """docstring here"""
        pass

    some_func.__doc__ = "new_docstring"

    assert "docstring here" not in some_func.__doc__
    assert "new_docstring" in some_func.__doc__
    assert "some_message" in some_func.__doc__


def test_deprecatd_docstring():
    message = "Use something else instead"

    @deprecated()
    def some_func():
        """This is a function"""

    @deprecated(message)
    def other_func():
        """This is another function"""

    assert ".. deprecated" in some_func.__doc__
    assert f".. deprecated\n   {message}" in other_func.__doc__


@pytest.fixture
def capture(request):
    handler = logbook.TestHandler(level=logbook.WARNING)
    handler.push_application()

    @request.addfinalizer
    def pop():
        handler.pop_application()

    return handler


@deprecated
def deprecated_func(a, b):
    return a + b


@pytest.fixture(autouse=True)
def forget_locations():
    forget_deprecation_locations()



import gzip
import os
import time
from datetime import datetime

import brotli
import pytest

import logbook

from .utils import LETTERS, capturing_stderr_context


def test_file_handler(logfile, activation_strategy, logger):
    handler = logbook.FileHandler(
        logfile,
        format_string="{record.level_name}:{record.channel}:{record.message}",
    )
    with activation_strategy(handler):
        logger.warn("warning message")
    handler.close()
    with open(logfile) as f:
        assert f.readline() == "WARNING:testlogger:warning message\n"


def test_file_handler_unicode(logfile, activation_strategy, logger):
    with capturing_stderr_context() as captured:
        with activation_strategy(logbook.FileHandler(logfile)):
            logger.info("\u0431")
    assert not captured.getvalue()


def test_file_handler_delay(logfile, activation_strategy, logger):
    handler = logbook.FileHandler(
        logfile,
        format_string="{record.level_name}:{record.channel}:{record.message}",
        delay=True,
    )
    assert not os.path.isfile(logfile)
    with activation_strategy(handler):
        logger.warn("warning message")
    handler.close()

    with open(logfile) as f:
        assert f.readline() == "WARNING:testlogger:warning message\n"


def test_monitoring_file_handler(logfile, activation_strategy, logger):
    if os.name == "nt":
        pytest.skip("unsupported on windows due to different IO (also unneeded)")
    handler = logbook.MonitoringFileHandler(
        logfile,
        format_string="{record.level_name}:{record.channel}:{record.message}",
        delay=True,
    )
    with activation_strategy(handler):
        logger.warn("warning message")
        os.rename(logfile, os.fspath(logfile) + ".old")
        logger.warn("another warning message")
    handler.close()
    with open(logfile) as f:
        assert f.read().strip() == "WARNING:testlogger:another warning message"


def test_custom_formatter(activation_strategy, logfile, logger):
    def custom_format(record, handler):
        return record.level_name + ":" + record.message

    handler = logbook.FileHandler(logfile)
    with activation_strategy(handler):
        handler.formatter = custom_format
        logger.warn("Custom formatters are awesome")

    with open(logfile) as f:
        assert f.readline() == "WARNING:Custom formatters are awesome\n"


def test_rotating_file_handler(logfile, activation_strategy, logger):
    basename = os.path.basename(logfile)
    handler = logbook.RotatingFileHandler(
        logfile,
        max_size=2048,
        backup_count=3,
    )
    handler.format_string = "{record.message}"
    with activation_strategy(handler):
        for c, x in zip(LETTERS, range(32)):
            logger.warn(c * 256)
    files = [x for x in os.listdir(os.path.dirname(logfile)) if x.startswith(basename)]
    files.sort()

    assert files == [basename, basename + ".1", basename + ".2", basename + ".3"]
    with open(logfile) as f:
        assert f.readline().rstrip() == ("C" * 256)
        assert f.readline().rstrip() == ("D" * 256)
        assert f.readline().rstrip() == ("E" * 256)
        assert f.readline().rstrip() == ("F" * 256)


@pytest.mark.parametrize("backup_count", [1, 3])
def test_timed_rotating_file_handler(tmpdir, activation_strategy, backup_count):
    basename = str(tmpdir.join("trot.log"))
    handler = logbook.TimedRotatingFileHandler(basename, backup_count=backup_count)
    handler.format_string = "[{record.time:%H:%M}] {record.message}"

    def fake_record(message, year, month, day, hour=0, minute=0, second=0):
        lr = logbook.LogRecord("Test Logger", logbook.WARNING, message)
        lr.time = datetime(year, month, day, hour, minute, second)
        return lr

    with activation_strategy(handler):
        for x in range(10):
            handler.handle(fake_record("First One", 2010, 1, 5, x + 1))
        for x in range(20):
            handler.handle(fake_record("Second One", 2010, 1, 6, x + 1))
        for x in range(10):
            handler.handle(fake_record("Third One", 2010, 1, 7, x + 1))
        for x in range(20):
            handler.handle(fake_record("Last One", 2010, 1, 8, x + 1))

    files = sorted(x for x in os.listdir(str(tmpdir)) if x.startswith("trot"))

    assert files == [f"trot-2010-01-0{i}.log" for i in range(5, 9)][-backup_count:]
    with open(str(tmpdir.join("trot-2010-01-08.log"))) as f:
        assert f.readline().rstrip() == "[01:00] Last One"
        assert f.readline().rstrip() == "[02:00] Last One"
    if backup_count > 1:
        with open(str(tmpdir.join("trot-2010-01-07.log"))) as f:
            assert f.readline().rstrip() == "[01:00] Third One"
            assert f.readline().rstrip() == "[02:00] Third One"


@pytest.mark.parametrize("backup_count", [1, 3])
def test_timed_rotating_file_handler__rollover_format(
    tmpdir, activation_strategy, backup_count
):
    basename = str(tmpdir.join("trot.log"))
    handler = logbook.TimedRotatingFileHandler(
        basename,
        backup_count=backup_count,
        rollover_format="{basename}{ext}.{timestamp}",
    )
    handler.format_string = "[{record.time:%H:%M}] {record.message}"

    def fake_record(message, year, month, day, hour=0, minute=0, second=0):
        lr = logbook.LogRecord("Test Logger", logbook.WARNING, message)
        lr.time = datetime(year, month, day, hour, minute, second)
        return lr

    with activation_strategy(handler):
        for x in range(10):
            handler.handle(fake_record("First One", 2010, 1, 5, x + 1))
        for x in range(20):
            handler.handle(fake_record("Second One", 2010, 1, 6, x + 1))
        for x in range(10):
            handler.handle(fake_record("Third One", 2010, 1, 7, x + 1))
        for x in range(20):
            handler.handle(fake_record("Last One", 2010, 1, 8, x + 1))

    files = sorted(x for x in os.listdir(str(tmpdir)) if x.startswith("trot"))

    assert files == [f"trot.log.2010-01-0{i}" for i in range(5, 9)][-backup_count:]
    with open(str(tmpdir.join("trot.log.2010-01-08"))) as f:
        assert f.readline().rstrip() == "[01:00] Last One"
        assert f.readline().rstrip() == "[02:00] Last One"
    if backup_count > 1:
        with open(str(tmpdir.join("trot.log.2010-01-07"))) as f:
            assert f.readline().rstrip() == "[01:00] Third One"
            assert f.readline().rstrip() == "[02:00] Third One"


@pytest.mark.parametrize("backup_count", [1, 3])
@pytest.mark.parametrize("preexisting_file", [True, False])
def test_timed_rotating_file_handler__not_timed_filename_for_current(
    tmpdir, activation_strategy, backup_count, preexisting_file
):
    basename = str(tmpdir.join("trot.log"))

    if preexisting_file:
        with open(basename, "w") as file:
            file.write("contents")
        jan_first = time.mktime(datetime(2010, 1, 1).timetuple())
        os.utime(basename, (jan_first, jan_first))

    handler = logbook.TimedRotatingFileHandler(
        basename,
        format_string="[{record.time:%H:%M}] {record.message}",
        backup_count=backup_count,
        rollover_format="{basename}{ext}.{timestamp}",
        timed_filename_for_current=False,
    )

    def fake_record(message, year, month, day, hour=0, minute=0, second=0):
        lr = logbook.LogRecord("Test Logger", logbook.WARNING, message)
        lr.time = datetime(year, month, day, hour, minute, second)
        return lr

    with activation_strategy(handler):
        for x in range(10):
            handler.handle(fake_record("First One", 2010, 1, 5, x + 1))
        for x in range(20):
            handler.handle(fake_record("Second One", 2010, 1, 6, x + 1))
        for x in range(10):
            handler.handle(fake_record("Third One", 2010, 1, 7, x + 1))
        for x in range(20):
            handler.handle(fake_record("Last One", 2010, 1, 8, x + 1))

    computed_files = [x for x in os.listdir(str(tmpdir)) if x.startswith("trot")]

    expected_files = ["trot.log.2010-01-01"] if preexisting_file else []
    expected_files += [f"trot.log.2010-01-0{i}" for i in range(5, 8)]
    expected_files += ["trot.log"]
    expected_files = expected_files[-backup_count:]

    assert sorted(computed_files) == sorted(expected_files)

    with open(str(tmpdir.join("trot.log"))) as f:
        assert f.readline().rstrip() == "[01:00] Last One"
        assert f.readline().rstrip() == "[02:00] Last One"
    if backup_count > 1:
        with open(str(tmpdir.join("trot.log.2010-01-07"))) as f:
            assert f.readline().rstrip() == "[01:00] Third One"
            assert f.readline().rstrip() == "[02:00] Third One"


def _decompress(input_file_name, use_gzip=True):
    if use_gzip:
        with gzip.open(input_file_name, "rb") as in_f:
            return in_f.read().decode()
    else:
        with open(input_file_name, "rb") as in_f:
            return brotli.decompress(in_f.read()).decode()


@pytest.mark.parametrize("use_gzip", [True, False])
def test_compression_file_handler(logfile, activation_strategy, logger, use_gzip):
    handler = (
        logbook.GZIPCompressionHandler(logfile)
        if use_gzip
        else logbook.BrotliCompressionHandler(logfile)
    )
    handler.format_string = "{record.level_name}:{record.channel}:{record.message}"
    with activation_strategy(handler):
        logger.warn("warning message")
    handler.close()
    assert _decompress(logfile, use_gzip) == "WARNING:testlogger:warning message\n"



"""Test with no handler active"""

from io import StringIO

from logbook import ERROR, Logger, NullHandler, StreamHandler

log = Logger("Test logger")


def run():
    out = StringIO()
    with NullHandler():
        with StreamHandler(out, level=ERROR) as handler:
            for x in range(500):
                log.warning("this is not handled")
    assert not out.getvalue()



"""Tests redirects from logging to logbook"""

from io import StringIO
from logging import StreamHandler, getLogger

from logbook.compat import LoggingHandler

log = getLogger("Test logger")


def run():
    out = StringIO()
    log.addHandler(StreamHandler(out))
    with LoggingHandler():
        for x in range(500):
            log.warning("this is not handled")
    assert out.getvalue().count("\n") == 500



"""Test with no handler active"""

from logging import getLogger

root_logger = getLogger()


def run():
    for x in range(500):
        getLogger("Test")
        del root_logger.manager.loggerDict["Test"]



import base64
import re
import ssl
from unittest.mock import ANY, call, patch

import logbook

from .utils import capturing_stderr_context, make_fake_mail_handler

__file_without_pyc__ = __file__
if __file_without_pyc__.endswith(".pyc"):
    __file_without_pyc__ = __file_without_pyc__[:-1]


def test_mail_handler(activation_strategy, logger):
    subject = "\xf8nicode"
    handler = make_fake_mail_handler(subject=subject)
    with capturing_stderr_context() as fallback:
        with activation_strategy(handler):
            logger.warn("This is not mailed")
            try:
                1 / 0
            except Exception:
                logger.exception("Viva la Espa\xf1a")

        if not handler.mails:
            # if sending the mail failed, the reason should be on stderr
            assert False, fallback.getvalue()

        assert len(handler.mails) == 1
        sender, receivers, mail = handler.mails[0]
        mail = mail.replace("\r", "")
        assert sender == handler.from_addr
        assert "=?utf-8?q?=C3=B8nicode?=" in mail
        header, data = mail.split("\n\n", 1)
        if "Content-Transfer-Encoding: base64" in header:
            data = base64.b64decode(data).decode("utf-8")
        assert re.search(r"Message type:\s+ERROR", data)
        assert re.search(r"Location:.*%s" % re.escape(__file_without_pyc__), data)
        assert re.search(r"Module:\s+%s" % __name__, data)
        assert re.search(r"Function:\s+test_mail_handler", data)
        body = "Viva la Espa\xf1a"
        assert body in data
        assert "\nTraceback (most" in data
        assert "1 / 0" in data
        assert "This is not mailed" in fallback.getvalue()


def test_mail_handler_batching(activation_strategy, logger):
    mail_handler = make_fake_mail_handler()
    handler = logbook.FingersCrossedHandler(mail_handler, reset=True)
    with activation_strategy(handler):
        logger.warn("Testing")
        logger.debug("Even more")
        logger.error("And this triggers it")
        logger.info("Aha")
        logger.error("And this triggers it again!")

    assert len(mail_handler.mails) == 2
    mail = mail_handler.mails[0][2]

    pieces = mail.split("Log records that led up to this one:")
    assert len(pieces) == 2
    body, rest = pieces
    rest = rest.replace("\r", "")

    assert re.search(r"Message type:\s+ERROR", body)
    assert re.search(r"Module:\s+%s" % __name__, body)
    assert re.search(r"Function:\s+test_mail_handler_batching", body)

    related = rest.strip().split("\n\n")
    assert len(related) == 2
    assert re.search(r"Message type:\s+WARNING", related[0])
    assert re.search(r"Message type:\s+DEBUG", related[1])

    assert "And this triggers it again" in mail_handler.mails[1][2]


def test_group_handler_mail_combo(activation_strategy, logger):
    mail_handler = make_fake_mail_handler(level=logbook.DEBUG)
    handler = logbook.GroupHandler(mail_handler)
    with activation_strategy(handler):
        logger.error("The other way round")
        logger.warn("Testing")
        logger.debug("Even more")
        assert mail_handler.mails == []

    assert len(mail_handler.mails) == 1
    mail = mail_handler.mails[0][2]

    pieces = mail.split("Other log records in the same group:")
    assert len(pieces) == 2
    body, rest = pieces
    rest = rest.replace("\r", "")

    assert re.search(r"Message type:\s+ERROR", body)
    assert re.search(r"Module:\s+" + __name__, body)
    assert re.search(r"Function:\s+test_group_handler_mail_combo", body)

    related = rest.strip().split("\n\n")
    assert len(related) == 2
    assert re.search(r"Message type:\s+WARNING", related[0])
    assert re.search(r"Message type:\s+DEBUG", related[1])


def test_mail_handler_arguments():
    patch_smtp = patch("smtplib.SMTP", autospec=True)
    patch_load_cert_chain = patch("ssl.SSLContext.load_cert_chain", autospec=True)

    with patch_load_cert_chain as mock_load_cert_chain:
        with patch_smtp as mock_smtp:
            # Test the mail handler with supported arguments before changes to
            # secure, credentials, and starttls
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr=("server.example.com", 465),
                credentials=("username", "password"),
                secure=("keyfile", "certfile"),
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("server.example.com", 465)
            mock_smtp().starttls.assert_called_once_with(context=ANY)
            assert isinstance(
                mock_smtp().starttls.call_args.kwargs["context"], ssl.SSLContext
            )
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_called_once_with("certfile", "keyfile")
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test secure=()
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr=("server.example.com", 465),
                credentials=("username", "password"),
                secure=(),
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("server.example.com", 465)
            mock_smtp().starttls.assert_called_once_with(context=None)
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_not_called()
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test implicit port with string server_addr, dictionary credentials,
            # dictionary secure.
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr="server.example.com",
                credentials={"user": "username", "password": "password"},
                secure={"certfile": "certfile2", "keyfile": "keyfile2"},
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("server.example.com", 587)
            mock_smtp().starttls.assert_called_once_with(context=ANY)
            assert isinstance(
                mock_smtp().starttls.call_args.kwargs["context"], ssl.SSLContext
            )
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_called_once_with("certfile2", "keyfile2")
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test default port for non-secure connection
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr="server.example.com",
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("server.example.com", 25)
            mock_smtp().starttls.assert_not_called()
            mock_load_cert_chain.assert_not_called()
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test default host and port for non-secure connection
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("127.0.0.1", 25)
            mock_smtp().starttls.assert_not_called()
            mock_load_cert_chain.assert_not_called()
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test default host and port for starttls connection
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                secure=True,
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("127.0.0.1", 587)
            mock_smtp().starttls.assert_called_once_with(context=None)
            mock_load_cert_chain.assert_not_called()
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test secure=True
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr=("server.example.com", 465),
                credentials=("username", "password"),
                secure=True,
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("server.example.com", 465)
            mock_smtp().starttls.assert_called_once_with(context=None)
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_not_called()
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test secure=False
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr=("server.example.com", 465),
                credentials=("username", "password"),
                secure=False,
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("server.example.com", 465)
            mock_smtp().starttls.assert_not_called()
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_not_called()
            mock_smtp.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test SSLContext
            context = ssl.create_default_context()
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr=("server.example.com", 465),
                credentials=("username", "password"),
                secure=context,
            )

            mail_handler.get_connection()

            mock_smtp.assert_called_once_with("server.example.com", 465)
            mock_smtp().starttls.assert_called_once_with(context=context)
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_not_called()

        with patch("smtplib.SMTP_SSL", autospec=True) as mock_smtp_ssl:
            # Test starttls=False
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr="server.example.com",
                credentials={"user": "username", "password": "password"},
                secure={"certfile": "certfile", "keyfile": "keyfile"},
                starttls=False,
            )

            mail_handler.get_connection()

            mock_smtp_ssl.assert_called_once_with(
                "server.example.com", 465, context=ANY
            )
            assert isinstance(mock_smtp_ssl.call_args.kwargs["context"], ssl.SSLContext)
            mock_load_cert_chain.assert_called_once_with("certfile", "keyfile")
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_smtp_ssl.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test starttls=False with secure=True
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr="server.example.com",
                credentials={"user": "username", "password": "password"},
                secure=True,
                starttls=False,
            )

            mail_handler.get_connection()

            mock_smtp_ssl.assert_called_once_with(
                "server.example.com", 465, context=None
            )
            mock_smtp_ssl().starttls.assert_not_called()
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_not_called()
            mock_smtp_ssl.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test default host and port for starttls connection
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                secure=True,
                starttls=False,
            )

            mail_handler.get_connection()

            mock_smtp_ssl.assert_called_once_with("127.0.0.1", 465, context=None)
            mock_smtp_ssl().starttls.assert_not_called()
            mock_load_cert_chain.assert_not_called()
            mock_smtp_ssl.reset_mock()
            mock_load_cert_chain.reset_mock()

            # Test SSLContext
            context = ssl.create_default_context()
            mail_handler = logbook.MailHandler(
                from_addr="from@example.com",
                recipients="to@example.com",
                server_addr="server.example.com",
                credentials={"user": "username", "password": "password"},
                secure=context,
                starttls=False,
            )

            mail_handler.get_connection()

            mock_smtp_ssl.assert_called_once_with(
                "server.example.com", 465, context=context
            )
            mock_smtp_ssl().starttls.assert_not_called()
            mock_smtp().login.assert_called_once_with("username", "password")
            mock_load_cert_chain.assert_not_called()



"""
    logbook
    ~~~~~~~

    Simple logging library that aims to support desktop, command line
    and web applications alike.

    :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
    :license: BSD, see LICENSE for more details.
"""

import os

from .base import (
    CRITICAL,
    DEBUG,
    ERROR,
    INFO,
    NOTICE,
    NOTSET,
    TRACE,
    WARNING,
    Flags,
    Logger,
    LoggerGroup,
    LogRecord,
    NestedSetup,
    Processor,
    dispatch_record,
    get_level_name,
    lookup_level,
    set_datetime_format,
)
from .handlers import (
    BrotliCompressionHandler,
    FileHandler,
    FingersCrossedHandler,
    GMailHandler,
    GroupHandler,
    GZIPCompressionHandler,
    Handler,
    HashingHandlerMixin,
    LimitingHandlerMixin,
    MailHandler,
    MonitoringFileHandler,
    NTEventLogHandler,
    NullHandler,
    RotatingFileHandler,
    StderrHandler,
    StreamHandler,
    StringFormatter,
    StringFormatterHandlerMixin,
    SyslogHandler,
    TestHandler,
    TimedRotatingFileHandler,
    WrapperHandler,
    create_syshandler,
)

from . import compat  # isort: skip

# create an anonymous default logger and provide all important
# methods of that logger as global functions
_default_logger = Logger("Generic")
_default_logger.suppress_dispatcher = True
trace = _default_logger.trace
debug = _default_logger.debug
info = _default_logger.info
warn = _default_logger.warn
warning = _default_logger.warning
notice = _default_logger.notice
error = _default_logger.error
exception = _default_logger.exception
catch_exceptions = _default_logger.catch_exceptions
critical = _default_logger.critical
log = _default_logger.log
del _default_logger


# install a default global handler
if os.environ.get("LOGBOOK_INSTALL_DEFAULT_HANDLER"):
    default_handler = StderrHandler()
    default_handler.push_application()

from .__version__ import __version__



import logbook

from .utils import capturing_stderr_context


def test_fingerscrossed(activation_strategy, logger, default_handler):
    handler = logbook.FingersCrossedHandler(default_handler, logbook.WARNING)

    # if no warning occurs, the infos are not logged
    with activation_strategy(handler):
        with capturing_stderr_context() as captured:
            logger.info("some info")
        assert captured.getvalue() == ""
        assert not handler.triggered

    # but if it does, all log messages are output
    with activation_strategy(handler):
        with capturing_stderr_context() as captured:
            logger.info("some info")
            logger.warning("something happened")
            logger.info("something else happened")
        logs = captured.getvalue()
        assert "some info" in logs
        assert "something happened" in logs
        assert "something else happened" in logs
        assert handler.triggered


def test_fingerscrossed_factory(activation_strategy, logger):
    handlers = []

    def handler_factory(record, fch):
        handler = logbook.TestHandler()
        handlers.append(handler)
        return handler

    def make_fch():
        return logbook.FingersCrossedHandler(handler_factory, logbook.WARNING)

    fch = make_fch()
    with activation_strategy(fch):
        logger.info("some info")
        assert len(handlers) == 0
        logger.warning("a warning")
        assert len(handlers) == 1
        logger.error("an error")
        assert len(handlers) == 1
        assert handlers[0].has_infos
        assert handlers[0].has_warnings
        assert handlers[0].has_errors
        assert not handlers[0].has_notices
        assert not handlers[0].has_criticals
        assert not handlers[0].has_debugs

    fch = make_fch()
    with activation_strategy(fch):
        logger.info("some info")
        logger.warning("a warning")
        assert len(handlers) == 2


def test_fingerscrossed_buffer_size(activation_strategy):
    logger = logbook.Logger("Test")
    test_handler = logbook.TestHandler()
    handler = logbook.FingersCrossedHandler(test_handler, buffer_size=3)

    with activation_strategy(handler):
        logger.info("Never gonna give you up")
        logger.warn("Aha!")
        logger.warn("Moar!")
        logger.error("Pure hate!")

    assert test_handler.formatted_records == [
        "[WARNING] Test: Aha!",
        "[WARNING] Test: Moar!",
        "[ERROR] Test: Pure hate!",
    ]



"""Tests with frame introspection disabled"""

from logbook import Flags, Logger, NullHandler

log = Logger("Test logger")


class DummyHandler(NullHandler):
    blackhole = False


def run():
    with Flags(introspection=False):
        with DummyHandler() as handler:
            for x in range(500):
                log.warning("this is not handled")



"""Tests basic stack manipulation performance"""

from io import StringIO
from tempfile import NamedTemporaryFile

from logbook import ERROR, WARNING, FileHandler, Handler, NullHandler, StreamHandler


def run():
    f = NamedTemporaryFile()
    out = StringIO()
    with NullHandler():
        with StreamHandler(out, level=WARNING):
            with FileHandler(f.name, level=ERROR):
                for x in range(100):
                    list(Handler.stack_manager.iter_context_objects())



import pytest

import logbook


def test_global_functions(activation_strategy):
    with activation_strategy(logbook.TestHandler()) as handler:
        logbook.debug("a debug message")
        logbook.info("an info message")
        logbook.warn("warning part 1")
        logbook.warning("warning part 2")
        logbook.notice("notice")
        logbook.error("an error")
        logbook.critical("pretty critical")
        logbook.log(logbook.CRITICAL, "critical too")

    assert handler.has_debug("a debug message")
    assert handler.has_info("an info message")
    assert handler.has_warning("warning part 1")
    assert handler.has_warning("warning part 2")
    assert handler.has_notice("notice")
    assert handler.has_error("an error")
    assert handler.has_critical("pretty critical")
    assert handler.has_critical("critical too")
    assert handler.records[0].channel == "Generic"
    assert handler.records[0].dispatcher is None


def test_level_lookup_failures():
    with pytest.raises(LookupError):
        logbook.get_level_name(37)
    with pytest.raises(LookupError):
        logbook.lookup_level("FOO")



from textwrap import dedent

import logbook

from .utils import make_fake_mail_handler


def test_handler_filter_after_processor(activation_strategy, logger):
    handler = make_fake_mail_handler(
        format_string=dedent(
            """
            Subject: Application Error for {record.extra[path]} [{record.extra[method]}]

            Message type:       {record.level_name}
            Location:           {record.filename}:{record.lineno}
            Module:             {record.module}
            Function:           {record.func_name}
            Time:               {record.time:%Y-%m-%d %H:%M:%S}
            Remote IP:          {record.extra[ip]}
            Request:            {record.extra[path]} [{record.extra[method]}]

            Message:

            {record.message}
            """
        ).lstrip(),
        filter=lambda r, h: "ip" in r.extra,
        bubble=False,
    )

    class Request:
        remote_addr = "127.0.0.1"
        method = "GET"
        path = "/index.html"

    def handle_request(request):
        def inject_extra(record):
            record.extra["ip"] = request.remote_addr
            record.extra["method"] = request.method
            record.extra["path"] = request.path

        processor = logbook.Processor(inject_extra)
        with activation_strategy(processor):
            handler.push_thread()
            try:
                try:
                    1 / 0
                except Exception:
                    logger.exception("Exception happened during request")
            finally:
                handler.pop_thread()

    handle_request(Request())
    assert len(handler.mails) == 1
    mail = handler.mails[0][2]
    assert "Subject: Application Error for /index.html [GET]" in mail
    assert "1 / 0" in mail


def test_handler_processors(activation_strategy, logger):
    handler = make_fake_mail_handler(
        format_string=dedent(
            """
            Subject: Application Error for {record.extra[path]} [{record.extra[method]}]

            Message type:       {record.level_name}
            Location:           {record.filename}:{record.lineno}
            Module:             {record.module}
            Function:           {record.func_name}
            Time:               {record.time:%Y-%m-%d %H:%M:%S}
            Remote IP:          {record.extra[ip]}
            Request:            {record.extra[path]} [{record.extra[method]}]

            Message:

            {record.message}
            """
        ).lstrip()
    )

    class Request:
        remote_addr = "127.0.0.1"
        method = "GET"
        path = "/index.html"

    def handle_request(request):
        def inject_extra(record):
            record.extra["ip"] = request.remote_addr
            record.extra["method"] = request.method
            record.extra["path"] = request.path

        processor = logbook.Processor(inject_extra)
        with activation_strategy(processor):
            handler.push_thread()
            try:
                try:
                    1 / 0
                except Exception:
                    logger.exception("Exception happened during request")
            finally:
                handler.pop_thread()

    handle_request(Request())
    assert len(handler.mails) == 1
    mail = handler.mails[0][2]
    assert "Subject: Application Error for /index.html [GET]" in mail
    assert "1 / 0" in mail



"""Benchmarks the file handler"""

from tempfile import NamedTemporaryFile

from logbook import FileHandler, Logger

log = Logger("Test logger")


def run():
    f = NamedTemporaryFile()
    with FileHandler(f.name) as handler:
        for x in range(500):
            log.warning("this is handled")



#!/usr/bin/env python
"""
    Runs the benchmarks
"""
import os
import re
import sys
from subprocess import Popen

try:
    from pkg_resources import get_distribution

    version = get_distribution("Logbook").version
except Exception:
    version = "unknown version"


_filename_re = re.compile(r"^bench_(.*?)\.py$")
bench_directory = os.path.abspath(os.path.dirname(__file__))


def list_benchmarks():
    result = []
    for name in os.listdir(bench_directory):
        match = _filename_re.match(name)
        if match is not None:
            result.append(match.group(1))
    result.sort(key=lambda x: (x.startswith("logging_"), x.lower()))
    return result


def run_bench(name, use_gevent=False):
    sys.stdout.write("%-32s" % name)
    sys.stdout.flush()
    Popen(
        [
            sys.executable,
            "-mtimeit",
            "-s",
            "from bench_%s import run" % name,
            "from logbook.concurrency import enable_gevent",
            "enable_gevent()" if use_gevent else "",
            "run()",
        ]
    ).wait()


def bench_wrapper(use_gevent=False):
    print("=" * 80)
    print(
        "Running benchmark with Logbook {} (gevent enabled={})".format(
            version, use_gevent
        )
    )
    print("-" * 80)
    os.chdir(bench_directory)
    for bench in list_benchmarks():
        run_bench(bench, use_gevent)
    print("-" * 80)


def main():
    bench_wrapper(False)
    try:
        import gevent

        bench_wrapper(True)
    except ImportError:
        pass


if __name__ == "__main__":
    main()



"""Benchmarks too low logger levels"""

from io import StringIO

from logbook import ERROR, Logger, StreamHandler

log = Logger("Test logger")
log.level = ERROR


def run():
    out = StringIO()
    with StreamHandler(out):
        for x in range(500):
            log.warning("this is not handled")



"""Like the filter test, but with the should_handle implemented"""

from io import StringIO

from logbook import Logger, NullHandler, StreamHandler

log = Logger("Test logger")


class CustomStreamHandler(StreamHandler):
    def should_handle(self, record):
        return False


def run():
    out = StringIO()
    with NullHandler():
        with CustomStreamHandler(out) as handler:
            for x in range(500):
                log.warning("this is not handled")
    assert not out.getvalue()



"""Tests the test handler"""

from logbook import Logger, TestHandler

log = Logger("Test logger")


def run():
    with TestHandler() as handler:
        for x in range(500):
            log.warning("this is not handled")



import re

import pytest


@pytest.mark.parametrize(
    "level, method",
    [
        ("trace", "has_traces"),
        ("debug", "has_debugs"),
        ("info", "has_infos"),
        ("notice", "has_notices"),
        ("warning", "has_warnings"),
        ("error", "has_errors"),
        ("critical", "has_criticals"),
    ],
)
def test_has_level(active_handler, logger, level, method):
    log = getattr(logger, level)
    log("Hello World")
    assert getattr(active_handler, method)


@pytest.mark.parametrize(
    "level, method",
    [
        ("trace", "has_trace"),
        ("debug", "has_debug"),
        ("info", "has_info"),
        ("notice", "has_notice"),
        ("warning", "has_warning"),
        ("error", "has_error"),
        ("critical", "has_critical"),
    ],
)
def test_regex_matching(active_handler, logger, level, method):
    log = getattr(logger, level)
    log("Hello World")
    has_level_method = getattr(active_handler, method)
    assert has_level_method(re.compile("^Hello"))
    assert not has_level_method(re.compile("world$"))
    assert not has_level_method("^Hello World")


def test_test_handler_cache(active_handler, logger):
    logger.warn("First line")
    assert len(active_handler.formatted_records) == 1
    # store cache, to make sure it is identifiable
    cache = active_handler.formatted_records
    assert len(active_handler.formatted_records) == 1
    assert cache is active_handler.formatted_records
    logger.warn("Second line invalidates cache")
    assert len(active_handler.formatted_records) == 2
    assert cache is not active_handler.formatted_records



"""Tests logging file handler in comparison"""

from logging import FileHandler, getLogger
from tempfile import NamedTemporaryFile

log = getLogger("Testlogger")


def run():
    f = NamedTemporaryFile()
    handler = FileHandler(f.name)
    log.addHandler(handler)
    for x in range(500):
        log.warning("this is handled \x6f")



import os

import pytest

import logbook

from .utils import require_module


@require_module("win32con")
@require_module("win32evtlog")
@require_module("win32evtlogutil")
@pytest.mark.skipif(
    os.environ.get("ENABLE_LOGBOOK_NTEVENTLOG_TESTS") is None,
    reason="Don't clutter NT Event Log unless enabled.",
)
def test_nteventlog_handler():
    from win32con import (
        EVENTLOG_ERROR_TYPE,
        EVENTLOG_INFORMATION_TYPE,
        EVENTLOG_WARNING_TYPE,
    )
    from win32evtlog import (
        EVENTLOG_BACKWARDS_READ,
        EVENTLOG_SEQUENTIAL_READ,
        OpenEventLog,
        ReadEventLog,
    )
    from win32evtlogutil import SafeFormatMessage

    logger = logbook.Logger("Test Logger")

    with logbook.NTEventLogHandler("Logbook Test Suite"):
        logger.info("The info log message.")
        logger.warning("The warning log message.")
        logger.error("The error log message.")

    def iter_event_log(handle, flags, offset):
        while True:
            events = ReadEventLog(handle, flags, offset)
            yield from events
            if not events:
                break

    handle = OpenEventLog(None, "Application")
    flags = EVENTLOG_BACKWARDS_READ | EVENTLOG_SEQUENTIAL_READ

    for event in iter_event_log(handle, flags, 0):
        source = str(event.SourceName)
        if source == "Logbook Test Suite":
            message = SafeFormatMessage(event, "Application")
            if "Message Level: INFO" in message:
                assert "The info log message" in message
                assert event.EventType == EVENTLOG_INFORMATION_TYPE
            if "Message Level: WARNING" in message:
                assert "The warning log message" in message
                assert event.EventType == EVENTLOG_WARNING_TYPE
            if "Message Level: ERROR" in message:
                assert "The error log message" in message
                assert event.EventType == EVENTLOG_ERROR_TYPE



import logbook

from .utils import capturing_stderr_context


def test_default_format_unicode(logger):
    with capturing_stderr_context() as stream:
        logger.warn("\u2603")
    assert "WARNING: testlogger: \u2603" in stream.getvalue()


def test_default_format_encoded(logger):
    with capturing_stderr_context() as stream:
        # it's a string but it's in the right encoding so don't barf
        logger.warn("\u2603")
    assert "WARNING: testlogger: \u2603" in stream.getvalue()


def test_default_format_bad_encoding(logger):
    with capturing_stderr_context() as stream:
        # it's a string, is wrong, but just dump it in the logger,
        # don't try to decode/encode it
        logger.warn("Русский".encode("koi8-r"))
    expected = "WARNING: testlogger: b'\\xf2\\xd5\\xd3\\xd3\\xcb\\xc9\\xca'"
    assert expected in stream.getvalue()


def test_custom_unicode_format_unicode(logger):
    format_string = "[{record.level_name}] {record.channel}: {record.message}"
    with capturing_stderr_context() as stream:
        with logbook.StderrHandler(format_string=format_string):
            logger.warn("\u2603")
    assert "[WARNING] testlogger: \u2603" in stream.getvalue()


def test_custom_string_format_unicode(logger):
    format_string = "[{record.level_name}] {record.channel}: {record.message}"
    with capturing_stderr_context() as stream:
        with logbook.StderrHandler(format_string=format_string):
            logger.warn("\u2603")
    assert "[WARNING] testlogger: \u2603" in stream.getvalue()


def test_unicode_message_encoded_params(logger):
    with capturing_stderr_context() as stream:
        logger.warn("\u2603 {0}", "\u2603".encode())
    assert "WARNING: testlogger: \u2603 b'\\xe2\\x98\\x83'" in stream.getvalue()


def test_encoded_message_unicode_params(logger):
    with capturing_stderr_context() as stream:
        logger.warn("\u2603 {0}".encode(), "\u2603")
    assert "WARNING: testlogger: \u2603 \u2603" in stream.getvalue()



"""Tests redirects from logging to logbook"""

from io import StringIO
from logging import getLogger

from logbook import StreamHandler
from logbook.compat import redirect_logging

redirect_logging()
log = getLogger("Test logger")


def run():
    out = StringIO()
    with StreamHandler(out):
        for x in range(500):
            log.warning("this is not handled")
    assert out.getvalue().count("\n") == 500



import os
import platform

from setuptools import Extension, setup

IS_CPYTHON = platform.python_implementation() == "CPython"
DISABLE_EXTENSION = bool(os.environ.get("DISABLE_LOGBOOK_CEXT"))


def status_msgs(*msgs):
    print("*" * 75)
    for msg in msgs:
        print(msg)
    print("*" * 75)


if not IS_CPYTHON:
    status_msgs(
        "WARNING: C extensions are not supported on this Python platform, "
        "speedups are not enabled.",
    )
    ext_modules = []
elif DISABLE_EXTENSION:
    status_msgs(
        "DISABLE_LOGBOOK_CEXT is set; not attempting to build C extensions.",
    )
    ext_modules = []
else:
    from Cython.Build import cythonize

    ext_modules = cythonize(
        [Extension("logbook._speedups", sources=["src/cython/speedups.pyx"])],
        language_level=3,
    )

setup(ext_modules=ext_modules)



"""
    logbook.more
    ~~~~~~~~~~~~

    Fancy stuff for logbook.

    :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
    :license: BSD, see LICENSE for more details.
"""

import os
import platform
import re
from collections import defaultdict
from functools import partial
from urllib.parse import parse_qsl, urlencode

from logbook._termcolors import colorize
from logbook.base import ERROR, NOTICE, NOTSET, RecordDispatcher, dispatch_record
from logbook.handlers import (
    Handler,
    StderrHandler,
    StringFormatter,
    StringFormatterHandlerMixin,
)
from logbook.ticketing import BackendBase
from logbook.ticketing import TicketingHandler as DatabaseHandler

try:
    import riemann_client.client
    import riemann_client.transport
except ImportError:
    riemann_client = None
    # from riemann_client.transport import TCPTransport, UDPTransport, BlankTransport

_ws_re = re.compile(r"(\s+)", re.UNICODE)
TWITTER_FORMAT_STRING = "[{record.channel}] {record.level_name}: {record.message}"
TWITTER_ACCESS_TOKEN_URL = "https://twitter.com/oauth/access_token"
NEW_TWEET_URL = "https://api.twitter.com/1/statuses/update.json"


class CouchDBBackend(BackendBase):
    """Implements a backend that writes into a CouchDB database."""

    def setup_backend(self):
        from couchdb import Server

        uri = self.options.pop("uri", "")
        couch = Server(uri)
        db_name = self.options.pop("db")
        self.database = couch[db_name]

    def record_ticket(self, record, data, hash, app_id):
        """Records a log record as ticket."""
        db = self.database

        ticket = record.to_dict()
        ticket["time"] = ticket["time"].isoformat() + "Z"
        ticket_id, _ = db.save(ticket)

        db.save(ticket)


class TwitterFormatter(StringFormatter):
    """Works like the standard string formatter and is used by the
    :class:`TwitterHandler` unless changed.
    """

    max_length = 140

    def format_exception(self, record):
        return f"{record.exception_shortname}: {record.exception_message}"

    def __call__(self, record, handler):
        formatted = StringFormatter.__call__(self, record, handler)
        rv = []
        length = 0
        for piece in _ws_re.split(formatted):
            length += len(piece)
            if length > self.max_length:
                if length - len(piece) < self.max_length:
                    rv.append("…")
                break
            rv.append(piece)
        return "".join(rv)


class TaggingLogger(RecordDispatcher):
    """A logger that attaches a tag to each record.  This is an alternative
    record dispatcher that does not use levels but tags to keep log
    records apart.  It is constructed with a descriptive name and at least
    one tag.  The tags are up for you to define::

        logger = TaggingLogger('My Logger', ['info', 'warning'])

    For each tag defined that way, a method appears on the logger with
    that name::

        logger.info('This is a info message')

    To dispatch to different handlers based on tags you can use the
    :class:`TaggingHandler`.

    The tags themselves are stored as list named ``'tags'`` in the
    :attr:`~logbook.LogRecord.extra` dictionary.
    """

    def __init__(self, name=None, tags=None):
        RecordDispatcher.__init__(self, name)
        # create a method for each tag named
        for tag in tags or ():
            setattr(self, tag, partial(self.log, tag))

    def log(self, tags, msg, *args, **kwargs):
        if isinstance(tags, str):
            tags = [tags]
        exc_info = kwargs.pop("exc_info", None)
        extra = kwargs.pop("extra", {})
        extra["tags"] = list(tags)
        frame_correction = kwargs.pop("frame_correction", 0)
        return self.make_record_and_handle(
            NOTSET, msg, args, kwargs, exc_info, extra, frame_correction
        )


class TaggingHandler(Handler):
    """A handler that logs for tags and dispatches based on those.

    Example::

        import logbook
        from logbook.more import TaggingHandler

        handler = TaggingHandler(dict(
            info=OneHandler(),
            warning=AnotherHandler()
        ))
    """

    def __init__(self, handlers, filter=None, bubble=False):
        Handler.__init__(self, NOTSET, filter, bubble)
        assert isinstance(handlers, dict)
        self._handlers = {
            tag: isinstance(handler, Handler) and [handler] or handler
            for (tag, handler) in handlers.items()
        }

    def emit(self, record):
        for tag in record.extra.get("tags", ()):
            for handler in self._handlers.get(tag, ()):
                handler.handle(record)


class TwitterHandler(Handler, StringFormatterHandlerMixin):
    """A handler that logs to twitter.  Requires that you sign up an
    application on twitter and request xauth support.  Furthermore the
    oauth2 library has to be installed.
    """

    default_format_string = TWITTER_FORMAT_STRING
    formatter_class = TwitterFormatter

    def __init__(
        self,
        consumer_key,
        consumer_secret,
        username,
        password,
        level=NOTSET,
        format_string=None,
        filter=None,
        bubble=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.username = username
        self.password = password

        try:
            import oauth2
        except ImportError:
            raise RuntimeError(
                "The python-oauth2 library is required for the TwitterHandler."
            )

        self._oauth = oauth2
        self._oauth_token = None
        self._oauth_token_secret = None
        self._consumer = oauth2.Consumer(consumer_key, consumer_secret)
        self._client = oauth2.Client(self._consumer)

    def get_oauth_token(self):
        """Returns the oauth access token."""
        if self._oauth_token is None:
            resp, content = self._client.request(
                TWITTER_ACCESS_TOKEN_URL + "?",
                "POST",
                body=urlencode(
                    {
                        "x_auth_username": self.username.encode("utf-8"),
                        "x_auth_password": self.password.encode("utf-8"),
                        "x_auth_mode": "client_auth",
                    }
                ),
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            if resp["status"] != "200":
                raise RuntimeError("unable to login to Twitter")
            data = dict(parse_qsl(content))
            self._oauth_token = data["oauth_token"]
            self._oauth_token_secret = data["oauth_token_secret"]
        return self._oauth.Token(self._oauth_token, self._oauth_token_secret)

    def make_client(self):
        """Creates a new oauth client auth a new access token."""
        return self._oauth.Client(self._consumer, self.get_oauth_token())

    def tweet(self, status):
        """Tweets a given status.  Status must not exceed 140 chars."""
        client = self.make_client()
        resp, content = client.request(
            NEW_TWEET_URL,
            "POST",
            body=urlencode({"status": status.encode("utf-8")}),
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        return resp["status"] == "200"

    def emit(self, record):
        self.tweet(self.format(record))


class SlackHandler(Handler, StringFormatterHandlerMixin):
    """A handler that logs to slack.  Requires that you sign up an
    application on slack and request an api token.  Furthermore the
    slacker library has to be installed.
    """

    def __init__(
        self,
        api_token,
        channel,
        level=NOTSET,
        format_string=None,
        filter=None,
        bubble=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)
        self.api_token = api_token

        try:
            from slacker import Slacker
        except ImportError:
            raise RuntimeError("The slacker library is required for the SlackHandler.")

        self.channel = channel
        self.slack = Slacker(api_token)

    def emit(self, record):
        self.slack.chat.post_message(channel=self.channel, text=self.format(record))


class JinjaFormatter:
    """A formatter object that makes it easy to format using a Jinja 2
    template instead of a format string.
    """

    def __init__(self, template):
        try:
            from jinja2 import Template
        except ImportError:
            raise RuntimeError("The jinja2 library is required for the JinjaFormatter.")
        self.template = Template(template)

    def __call__(self, record, handler):
        return self.template.render(record=record, handler=handler)


class ExternalApplicationHandler(Handler):
    """This handler invokes an external application to send parts of
    the log record to.  The constructor takes a list of arguments that
    are passed to another application where each of the arguments is a
    format string, and optionally a format string for data that is
    passed to stdin.

    For example it can be used to invoke the ``say`` command on OS X::

        from logbook.more import ExternalApplicationHandler
        say_handler = ExternalApplicationHandler(['say', '{record.message}'])

    Note that the above example is blocking until ``say`` finished, so it's
    recommended to combine this handler with the
    :class:`logbook.ThreadedWrapperHandler` to move the execution into
    a background thread.

    .. versionadded:: 0.3
    """

    def __init__(
        self,
        arguments,
        stdin_format=None,
        encoding="utf-8",
        level=NOTSET,
        filter=None,
        bubble=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        self.encoding = encoding
        self._arguments = list(arguments)
        if stdin_format is not None:
            stdin_format = stdin_format
        self._stdin_format = stdin_format
        import subprocess

        self._subprocess = subprocess

    def emit(self, record):
        args = [arg.format(record=record) for arg in self._arguments]
        if self._stdin_format is not None:
            stdin_data = self._stdin_format.format(record=record).encode(self.encoding)
            stdin = self._subprocess.PIPE
        else:
            stdin = None
        c = self._subprocess.Popen(args, stdin=stdin)
        if stdin is not None:
            c.communicate(stdin_data)
        c.wait()


class ColorizingStreamHandlerMixin:
    """A mixin class that does colorizing.

    .. versionadded:: 0.3
    .. versionchanged:: 1.0.0
       Added Windows support if `colorama`_ is installed.

    .. _`colorama`: https://pypi.org/project/colorama
    """

    _use_color = None

    def force_color(self):
        """Force colorizing the stream (`should_colorize` will return True)"""
        self._use_color = True

    def forbid_color(self):
        """Forbid colorizing the stream (`should_colorize` will return False)"""
        self._use_color = False

    def should_colorize(self, record):
        """Returns `True` if colorizing should be applied to this
        record.  The default implementation returns `True` if the
        stream is a tty. If we are executing on Windows, colorama must be
        installed.
        """
        if os.name == "nt":
            try:
                import colorama
            except ImportError:
                return False
        if self._use_color is not None:
            return self._use_color
        isatty = getattr(self.stream, "isatty", None)
        return isatty and isatty()

    def get_color(self, record):
        """Returns the color for this record."""
        if record.level >= ERROR:
            return "red"
        elif record.level >= NOTICE:
            return "yellow"
        return "lightgray"

    def format(self, record):
        rv = super().format(record)
        if self.should_colorize(record):
            if color := self.get_color(record):
                rv = colorize(color, rv)
        return rv


class ColorizedStderrHandler(ColorizingStreamHandlerMixin, StderrHandler):
    """A colorizing stream handler that writes to stderr.  It will only
    colorize if a terminal was detected.  Note that this handler does
    not colorize on Windows systems.

    .. versionadded:: 0.3
    .. versionchanged:: 1.0
       Added Windows support if `colorama`_ is installed.

    .. _`colorama`: https://pypi.org/project/colorama
    """

    def __init__(self, *args, **kwargs):
        StderrHandler.__init__(self, *args, **kwargs)

        # Try import colorama so that we work on Windows. colorama.init is a
        # noop on other operating systems.
        try:
            import colorama
        except ImportError:
            pass
        else:
            colorama.init()


# backwards compat.  Should go away in some future releases
from logbook.handlers import FingersCrossedHandler as FingersCrossedHandlerBase


class FingersCrossedHandler(FingersCrossedHandlerBase):
    def __init__(self, *args, **kwargs):
        FingersCrossedHandlerBase.__init__(self, *args, **kwargs)
        from warnings import warn

        warn(
            PendingDeprecationWarning(
                "fingers crossed handler changed "
                "location.  It's now a core component of Logbook."
            )
        )


class ExceptionHandler(Handler, StringFormatterHandlerMixin):
    """An exception handler which raises exceptions of the given `exc_type`.
    This is especially useful if you set a specific error `level` e.g. to treat
    warnings as exceptions::

        from logbook.more import ExceptionHandler

        class ApplicationWarning(Exception):
            pass

        exc_handler = ExceptionHandler(ApplicationWarning, level='WARNING')

    .. versionadded:: 0.3
    """

    def __init__(
        self, exc_type, level=NOTSET, format_string=None, filter=None, bubble=False
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)
        self.exc_type = exc_type

    def handle(self, record):
        if self.should_handle(record):
            raise self.exc_type(self.format(record))
        return False


class DedupHandler(Handler):
    """A handler that deduplicates log messages.

    It emits each unique log record once, along with the number of times it was
    emitted.
    Example:::

        with logbook.more.DedupHandler():
            logbook.error('foo')
            logbook.error('bar')
            logbook.error('foo')

    The expected output:::

       message repeated 2 times: foo
       message repeated 1 times: bar
    """

    def __init__(
        self, format_string="message repeated {count} times: {message}", *args, **kwargs
    ):
        Handler.__init__(self, bubble=False, *args, **kwargs)
        self._format_string = format_string
        self.clear()

    def clear(self):
        self._message_to_count = defaultdict(int)
        self._unique_ordered_records = []

    def pop_application(self):
        Handler.pop_application(self)
        self.flush()

    def pop_thread(self):
        Handler.pop_thread(self)
        self.flush()

    def pop_context(self):
        Handler.pop_context(self)
        self.flush()

    def pop_greenlet(self):
        Handler.pop_greenlet(self)
        self.flush()

    def handle(self, record):
        if record.message not in self._message_to_count:
            self._unique_ordered_records.append(record)
        self._message_to_count[record.message] += 1
        return True

    def flush(self):
        for record in self._unique_ordered_records:
            record.message = self._format_string.format(
                message=record.message, count=self._message_to_count[record.message]
            )
            # record.dispatcher is the logger who created the message,
            # it's sometimes supressed (by logbook.info for example)
            if record.dispatcher is not None:
                dispatch = record.dispatcher.call_handlers
            else:
                dispatch = dispatch_record
            dispatch(record)
        self.clear()


class RiemannHandler(Handler):
    """
    A handler that sends logs as events to Riemann.
    """

    def __init__(
        self,
        host,
        port,
        message_type="tcp",
        ttl=60,
        flush_threshold=10,
        bubble=False,
        filter=None,
        level=NOTSET,
    ):
        """
        :param host: riemann host
        :param port: riemann port
        :param message_type: selects transport. Currently available 'tcp' and 'udp'
        :param ttl: defines time to live in riemann
        :param flush_threshold: count of events after which we send to riemann
        """
        if riemann_client is None:
            raise NotImplementedError(
                "The Riemann handler requires the riemann_client package"
            )  # pragma: no cover
        Handler.__init__(self, level, filter, bubble)
        self.host = host
        self.port = port
        self.ttl = ttl
        self.queue = []
        self.flush_threshold = flush_threshold
        if message_type == "tcp":
            self.transport = riemann_client.transport.TCPTransport
        elif message_type == "udp":
            self.transport = riemann_client.transport.UDPTransport
        elif message_type == "test":
            self.transport = riemann_client.transport.BlankTransport
        else:
            msg = "Currently supported message types for RiemannHandler are: {}. \
                    {} is not supported.".format(
                ",".join(["tcp", "udp", "test"]), message_type
            )
            raise RuntimeError(msg)

    def record_to_event(self, record):
        from time import time

        tags = ["log", record.level_name]
        msg = str(record.exc_info[1]) if record.exc_info else record.msg
        channel_name = str(record.channel) if record.channel else "unknown"
        if any([record.level_name == keywords for keywords in ["ERROR", "EXCEPTION"]]):
            state = "error"
        else:
            state = "ok"
        return {
            "metric_f": 1.0,
            "tags": tags,
            "description": msg,
            "time": int(time()),
            "ttl": self.ttl,
            "host": platform.node(),
            "service": f"{channel_name}.{os.getpid()}",
            "state": state,
        }

    def _flush_events(self):
        with riemann_client.client.QueuedClient(
            self.transport(self.host, self.port)
        ) as cl:
            for event in self.queue:
                cl.event(**event)
            cl.flush()
        self.queue = []

    def emit(self, record):
        self.queue.append(self.record_to_event(record))

        if len(self.queue) == self.flush_threshold:
            self._flush_events()



from io import StringIO

from logbook import Logger, NullHandler, StreamHandler

log = Logger("Test logger")


def run():
    out = StringIO()
    with NullHandler():
        with StreamHandler(out, filter=lambda r, h: False) as handler:
            for x in range(500):
                log.warning("this is not handled")
    assert not out.getvalue()



"""Tests the stream handler"""

from io import StringIO

from logbook import Logger, StreamHandler

log = Logger("Test logger")


def run():
    out = StringIO()
    with StreamHandler(out) as handler:
        for x in range(500):
            log.warning("this is not handled")
    assert out.getvalue().count("\n") == 500
